import { Logger } from '$lib/log';
import type { OauthResponse } from '$lib/mastodon/response';
import fs from 'fs/promises';

export abstract class OauthPlaylistAdder {
  /// How many minutes before expiry the token will be refreshed
  protected refresh_time: number = 15;
  protected logger: Logger = new Logger('OauthPlaylistAdder');

  protected constructor(
    protected apiBase: string,
    protected token_file_name: string
  ) {}

  public async authCodeExists(): Promise<boolean> {
    try {
      const token = await this.auth();
      return token !== null && !token.error;
    } catch {
      this.logger.info('No auth token yet, authorizing...');
      return false;
    }
  }

  protected constructAuthUrlInternal(
    endpointUrl: string,
    clientId: string,
    scope: string,
    redirectUri: URL,
    additionalParameters: Map<string, string> = new Map()
  ): URL {
    const authUrl = new URL(endpointUrl);
    authUrl.searchParams.append('client_id', clientId);
    authUrl.searchParams.append('redirect_uri', redirectUri.toString());
    authUrl.searchParams.append('response_type', 'code');
    authUrl.searchParams.append('scope', scope);
    for (let p of additionalParameters.entries()) {
      authUrl.searchParams.append(p[0], p[1]);
    }
    return authUrl;
  }

  public async receivedAuthCodeInternal(
    tokenUrl: URL,
    clientId: string,
    code: string,
    redirectUri: URL,
    client_secret?: string,
    customHeader?: HeadersInit
  ) {
    this.logger.debug('received code');
    const params = new URLSearchParams();
    params.append('client_id', clientId);
    params.append('code', code);
    params.append('grant_type', 'authorization_code');
    params.append('redirect_uri', redirectUri.toString());
    if (client_secret) {
      params.append('client_secret', client_secret);
    }
    this.logger.debug('sending token req', params);
    const resp: OauthResponse = await fetch(tokenUrl, {
      method: 'POST',
      body: params,
      headers: customHeader
    }).then((r) => r.json());
    this.logger.debug('received access token', resp);
    let expiration = new Date();
    expiration.setTime(expiration.getTime() + resp.expires_in * 1000);
    expiration.setSeconds(expiration.getSeconds() + resp.expires_in);
    resp.expires = expiration;
    await fs.writeFile(this.token_file_name, JSON.stringify(resp));
  }

  protected async auth(): Promise<OauthResponse | null> {
    try {
      const token_file = await fs.readFile(this.token_file_name, { encoding: 'utf8' });
      let token = JSON.parse(token_file);
      if (token.expires) {
        if (typeof token.expires === typeof '') {
          token.expires = new Date(token.expires);
        }
      }
      return token;
    } catch (e) {
      this.logger.error('Could not read access token', e);
      return null;
    }
  }

  protected async shouldRefreshToken(): Promise<{ token: OauthResponse; refresh: boolean } | null> {
    const token = await this.auth();
    if (token == null || !token?.expires) {
      return null;
    }
    let refreshAt = new Date();
    refreshAt.setTime(refreshAt.getTime() - this.refresh_time * 60 * 1000);
    this.logger.info('token expiry', token.expires, 'vs refresh @', refreshAt);
    if (token.expires.getTime() > refreshAt.getTime()) {
      return {
        token: token,
        refresh: false
      };
    }

    this.logger.info(
      'Token expires',
      token.expires,
      token.expires.getTime(),
      `which is after the refresh time`,
      refreshAt,
      refreshAt.getTime()
    );
    return {
      token: token,
      refresh: true
    };
  }

  protected async requestRefreshToken(
    tokenUrl: URL,
    clientId: string,
    refresh_token: string,
    redirect_uri?: string,
    client_secret?: string,
    customHeader?: HeadersInit
  ) {
    const params = new URLSearchParams();
    params.append('client_id', clientId);
    params.append('grant_type', 'refresh_token');
    params.append('refresh_token', refresh_token);
    if (client_secret) {
      params.append('client_secret', client_secret);
    }
    if (redirect_uri) {
      params.append('redirect_uri', redirect_uri);
    }
    this.logger.debug('sending token req', params);
    const resp: OauthResponse = await fetch(tokenUrl, {
      method: 'POST',
      body: params,
      headers: customHeader
    }).then((r) => r.json());
    this.logger.verbose('received access token', resp);
    if (resp.error) {
      this.logger.error('token resp error', resp);
      return null;
    }
    if (!resp.refresh_token) {
      resp.refresh_token = refresh_token;
    }
    let expiration = new Date();
    expiration.setTime(expiration.getTime() + resp.expires_in * 1000);
    expiration.setSeconds(expiration.getSeconds() + resp.expires_in);
    resp.expires = expiration;
    await fs.writeFile(this.token_file_name, JSON.stringify(resp));
    return resp;
  }
}
